/*
 * Copyright (c) 2005, John Mettraux, OpenWFE.org
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * . Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.  
 * 
 * . Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * 
 * . Neither the name of the "OpenWFE" nor the names of its contributors may be
 *   used to endorse or promote products derived from this software without
 *   specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: RmiWorkSession.java 2356 2006-02-27 08:25:11Z jmettraux $
 */

//
// RmiWorkSession.java
//
// jmettraux@openwfe.org
//
// generated with 
// jtmpl 1.0.04 31.10.2002 John Mettraux (jmettraux@openwfe.org)
//

package openwfe.org.worklist.impl;

import java.rmi.RemoteException;
import java.security.PermissionCollection;
import java.security.PrivilegedActionException;
import javax.security.auth.Subject;

import openwfe.org.Service;
import openwfe.org.OpenWfeException;
import openwfe.org.ServiceException;
import openwfe.org.ApplicationContext;
import openwfe.org.rmi.session.ClassedWorkSessionServer;
import openwfe.org.auth.PolicyService;
import openwfe.org.engine.workitem.LaunchItem;
import openwfe.org.engine.workitem.InFlowWorkItem;
import openwfe.org.engine.expressions.FlowExpressionId;
import openwfe.org.worklist.Header;
import openwfe.org.worklist.Launchable;
import openwfe.org.worklist.WorkSession;
import openwfe.org.worklist.Definitions;
import openwfe.org.worklist.store.WorkItemStore;
import openwfe.org.worklist.LaunchException;
import openwfe.org.worklist.WorkListException;
import openwfe.org.worklist.auth.LaunchPermission;
import openwfe.org.worklist.auth.DelegationPermission;


/**
 * An implementation of WorkSession based on RMI.
 *
 * <p><font size=2>CVS Info :
 * <br>$Author: jmettraux $
 * <br>$Date: 2006-02-27 09:25:11 +0100 (Mon, 27 Feb 2006) $
 * <br>$Id: RmiWorkSession.java 2356 2006-02-27 08:25:11Z jmettraux $ </font>
 *
 * @author john.mettraux@openwfe.org
 */
public class RmiWorkSession

    extends java.rmi.server.UnicastRemoteObject

    implements WorkSession, Service

{

    private final static org.apache.log4j.Logger log = org.apache.log4j.Logger
        .getLogger(RmiWorkSession.class.getName());

    //
    // FIELDS

    private String serviceName = null;
    private ApplicationContext applicationContext = null;
    private java.util.Map serviceParams = null;
    private Subject subject = null;

    private java.util.Map ids = new java.util.HashMap();

    //
    // CONSTRUCTORS

    /*
     * This constructor must be present to rethrow its super generated 
     * RemoteException.
     */
    public RmiWorkSession ()
        throws RemoteException
    {
        super();
    }

    //
    // METHODS from WorkSession

    //
    // non remote methods

    public void init 
        (final String serviceName,
         final ApplicationContext context, 
         final java.util.Map params)
         //final Object subject)
    throws 
        //RemoteException
        ServiceException
    {
        log.debug("init()");

        this.serviceName = serviceName;
        this.applicationContext = context;
        this.serviceParams = params;

        this.subject = 
            (Subject)params.get(ClassedWorkSessionServer.P_SUBJECT);

        if (this.subject == null) 
            log.warn("init() no subject passed.");
    }

    public String getName ()
    {
        return this.serviceName;
    }

    public Subject getSubject () 
    { 
        return this.subject; 
    }

    public ApplicationContext getContext ()
    {
        return this.applicationContext;
    }

    public java.util.Map getParams ()
    {
        return this.serviceParams;
    }

    public org.jdom.Element getStatus ()
    {
        return null; // no need to implement it now
    }

    //
    // remote methods
    //
    // (a lot of security wrapping)

    /*
     * This first method is unprotected : everybody (if authenticated)
     * may list the storeNames
     */
    public java.util.List getStoreNames ()
        throws RemoteException
    {
        final java.util.List result = new java.util.LinkedList();

        final java.util.Iterator it = 
            this.applicationContext.keySet().iterator();
        while (it.hasNext())
        {
            final String serviceName = (String)it.next();

            final Object o = this.applicationContext.lookup(serviceName);

            if (o instanceof WorkItemStore) result.add(serviceName);
        }

        log.debug("getStorenames() found "+result.size()+" stores.");

        return result;
    }

    public int countWorkItems (final String storeName)
        throws RemoteException, WorkListException
    {
        try
        {
            Object result = Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.CountWorkItemsAction
                    (this.applicationContext,
                     this.subject,
                     storeName));

            return ((Integer)result).intValue();
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public java.util.List findFlowInstance
        (final String storeName, final String workflowInstanceId)
    throws 
        RemoteException, WorkListException
    {
        try
        {
            Object result = Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.FindFlowInstanceAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     workflowInstanceId));

            final java.util.List feis = (java.util.List)result;

            registerIds(feis);

            return feis;
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public java.util.List getHeaders 
        (final String storeName, final int limit)
    throws 
        RemoteException, WorkListException
    {
        try
        {
            Object result = Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.GetHeadersAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     limit,
                     null));

            final java.util.List headers = (java.util.List)result;

            registerIds(headers);

            return headers;
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public java.util.List getHeaders 
        (final String storeName, 
         final int limit, 
         final java.util.Comparator headerComparator)
    throws 
        RemoteException, WorkListException
    {
        try
        {
            Object result = Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.GetHeadersAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     limit,
                     headerComparator));

            return (java.util.List)result;
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public InFlowWorkItem get 
        (final String storeName, final FlowExpressionId fei)
    throws 
        RemoteException, WorkListException
    {
        try
        {
            Object result = Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.GetAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     fei));

            return (InFlowWorkItem)result;
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public InFlowWorkItem get 
        (final String storeName, final String id)
    throws 
        RemoteException, WorkListException
    {
        final FlowExpressionId fei = (FlowExpressionId)this.ids.get(id);

        return get(storeName, fei);
    }

    public InFlowWorkItem getAndLock 
        (final String storeName, final FlowExpressionId fei)
    throws 
        RemoteException, WorkListException
    {
        log.debug("getAndLock() storeName >"+storeName+"< fei "+fei);

        try
        {
            Object result = Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.GetAndLockAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     fei));

            return (InFlowWorkItem)result;
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public InFlowWorkItem getAndLock 
        (final String storeName, final String id)
    throws 
        RemoteException, WorkListException
    {
        final FlowExpressionId fei = (FlowExpressionId)this.ids.get(id);

        return getAndLock(storeName, fei);
    }

    public void release 
        (final String storeName, final InFlowWorkItem item)
    throws 
        RemoteException, WorkListException
    {
        try
        {
            Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.ReleaseAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     item));
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public void save 
        (final String storeName, final InFlowWorkItem item)
    throws 
        RemoteException, WorkListException
    {
        try
        {
            Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.SaveAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     item));
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public void forward 
        (final String storeName, final InFlowWorkItem item)
    throws 
        RemoteException, OpenWfeException
    {
        try
        {
            Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.ForwardAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     item));
        }
        catch (PrivilegedActionException pae)
        {
            log.debug("forward() failure", pae);

            throw (OpenWfeException)pae.getException();
        }
    }

    /**
     * This method returns a list of Launchable instances, ie objects
     * pointing to workflow definitions the user is authorized to launch.
     */
    public java.util.List getLaunchables ()
        throws RemoteException, WorkListException
    {
        java.util.List launchPermissions = 
            getPermissions(LaunchPermission.class);

        //
        // build launchables out of launch permissions
        
        java.util.Map result = new java.util.HashMap();

        java.util.Iterator it = launchPermissions.iterator();
        while (it.hasNext())
        {
            LaunchPermission lp = (LaunchPermission)it.next();
            java.util.Iterator lit = lp.resolveUrl().iterator();
            while (lit.hasNext())
            {
                String url = null;
                try
                {
                    url = (String)lit.next();

                    if (url.indexOf(".*") > -1) continue;

                    result.put
                        (url, 
                         new Launchable
                            (this.applicationContext, 
                             this.serviceParams, 
                             lp.getEngineId(), 
                             url));
                }
                catch (final LaunchException le)
                {
                    log.info
                        ("Failed to build launchable for "+
                         lp.getEngineId()+"::"+url);
                }
            }
        }

        //
        // return result
        
        java.util.ArrayList keys = new java.util.ArrayList(result.keySet());
        java.util.Collections.sort(keys);

        java.util.ArrayList launchables = new java.util.ArrayList(keys.size());

        it = keys.iterator();

        while (it.hasNext()) 
            launchables.add(result.get(it.next()));

        return launchables;
    }

    public String launch (final String engineId, final LaunchItem item)
        throws RemoteException, OpenWfeException
    {
        try
        {
            final Object result = Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.LaunchAction
                    (this.applicationContext,
                     this.subject,
                     engineId,
                     item));

            //log.debug
            //    ("launch() result is of class "+result.getClass().getName());

            return (String)result;
        }
        catch (PrivilegedActionException pae)
        {
            throw (OpenWfeException)pae.getException();
        }
    }

    public void delegate 
        (String storeName, InFlowWorkItem item, String targetStoreName)
    throws RemoteException, WorkListException
    {
        try
        {
            Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.DelegateAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     item,
                     targetStoreName));
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public java.util.List getParticipantsForDelegation ()
        throws RemoteException, WorkListException
    {
        java.util.List delegationPermissions = 
            getPermissions(DelegationPermission.class);

        java.util.List result = 
            new java.util.ArrayList(delegationPermissions.size());

        java.util.Iterator it = delegationPermissions.iterator();
        while (it.hasNext())
        {
            DelegationPermission dp = (DelegationPermission)it.next();
            result.add(dp.getName());
        }

        return result;
    }

    public void delegateToParticipant
        (String storeName, InFlowWorkItem item, String participantName)
    throws 
        RemoteException, WorkListException
    {
        try
        {
            Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.DelegateToParticipantAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     item,
                     participantName));
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }

    public boolean checkPermission (String storeName, String action)
        throws RemoteException, WorkListException
    {
        log.debug("checkPermission() action >"+action+"<");
        try
        {
            return ((Boolean)Subject.doAs
                (this.subject, 
                 new openwfe.org.wlactions.CheckPermissionAction
                    (this.applicationContext,
                     this.subject,
                     storeName,
                     action))).booleanValue();
        }
        catch (PrivilegedActionException pae)
        {
            throw (WorkListException)pae.getException();
        }
    }
    
    /**
     * Terminates this worksession.
     */
    public void close ()
        throws RemoteException
    {
        //
        // TODO : remove all locks !
    }

    //
    // OTHER METHODS

    /*
     * Keeps a map of the ' flow expression ids with the 
     * string id as key and with the workitem as values.
     */
    private void registerIds (final java.util.List l)
    {
        final java.util.Iterator it = l.iterator();
        while (it.hasNext())
        {
            final Object o = it.next();

            FlowExpressionId fei = null;

            if (o instanceof Header)
            {
                fei = ((Header)o).getExpressionId();
            }
            else if (o instanceof FlowExpressionId)
            {
                fei = (FlowExpressionId)o;
            }
            else
            {
                throw new IllegalArgumentException
                    ("Couldn't register flow expression id "+
                     "out of object of class "+o.getClass().getName());
            }

            this.ids.put(fei.asStringId(), fei);
        }
    }

    protected java.util.List getPermissions (Class permissionClass)
        throws WorkListException
    {
        //
        // get policy service
        
        PolicyService policy = 
            PolicyService.lookupPolicyService(this.applicationContext);
        policy.refresh();

        //
        // get permissions
        
        PermissionCollection permissions = null;
        try
        {
            permissions = policy.getPermissions(this.subject);
        }
        catch (openwfe.org.auth.AuthException ae)
        {
            throw new WorkListException
                ("Failed to list permissions for subject", ae);
        }

        //
        // take out valid permissions
        
        java.util.List result = new java.util.ArrayList();

        java.util.Enumeration en = permissions.elements();
        while (en.hasMoreElements())
        {
            Object o = en.nextElement();

            if (permissionClass.isAssignableFrom(o.getClass())) 
                result.add(o);
        }

        //
        // return result
        
        return result;
    }

}
